package solutions.bitbadger.documents.common

import kotlinx.serialization.json.Json
import java.sql.Connection
import java.sql.DriverManager

object Configuration {

    /**
     * JSON serializer; replace to configure with non-default options
     *
     * The default sets `encodeDefaults` to `true` and `explicitNulls` to `false`; see
     * https://github.com/Kotlin/kotlinx.serialization/blob/master/docs/json.md for all configuration options
     */
    var json = Json {
        encodeDefaults = true
        explicitNulls  = false
    }

    /** The field in which a document's ID is stored */
    var idField = "id"

    /** The automatic ID strategy to use */
    var autoIdStrategy = AutoId.DISABLED

    /** The length of automatic random hex character string */
    var idStringLength = 16

    /** The connection string for the JDBC connection */
    var connectionString: String? = null

    /**
     * Retrieve a new connection to the configured database
     *
     * @return A new connection to the configured database
     * @throws IllegalArgumentException If the connection string is not set before calling this
     */
    fun dbConn(): Connection {
        if (connectionString == null) {
            throw IllegalArgumentException("Please provide a connection string before attempting data access")
        }
        return DriverManager.getConnection(connectionString)
    }

    private var dialectValue: Dialect? = null

    /** The dialect in use */
    val dialect: Dialect
        get() {
            if (dialectValue == null) {
                if (connectionString == null) {
                    throw IllegalArgumentException("Please provide a connection string before attempting data access")
                }
                val it = connectionString!!
                dialectValue = when {
                    it.contains("sqlite") -> Dialect.SQLITE
                    it.contains("postgresql") -> Dialect.POSTGRESQL
                    else -> throw IllegalArgumentException("Cannot determine dialect from [$it]")
                }
            }

            return dialectValue!!
        }
}
