package solutions.bitbadger.documents.common

import java.sql.Connection
import java.sql.ResultSet

/**
 * Execute a query that returns a list of results
 *
 * @param query The query to retrieve the results
 * @param parameters Parameters to use for the query
 * @param mapFunc The mapping function between the document and the domain item
 * @return A list of results for the given query
 */
inline fun <reified TDoc> Connection.customList(query: String, parameters: Collection<Parameter<*>>,
                                          mapFunc: (ResultSet) -> TDoc): List<TDoc> =
    Parameters.apply(this, query, parameters).use { stmt ->
        Results.toCustomList(stmt, mapFunc)
    }

/**
 * Execute a query that returns one or no results
 *
 * @param query The query to retrieve the results
 * @param parameters Parameters to use for the query
 * @param mapFunc The mapping function between the document and the domain item
 * @return The document if one matches the query, `null` otherwise
 */
inline fun <reified TDoc> Connection.customSingle(query: String, parameters: Collection<Parameter<*>>,
                                                  mapFunc: (ResultSet) -> TDoc): TDoc? =
    this.customList("$query LIMIT 1", parameters, mapFunc).singleOrNull()

/**
 * Execute a query that returns no results
 *
 * @param query The query to retrieve the results
 * @param parameters Parameters to use for the query
 */
fun Connection.customNonQuery(query: String, parameters: Collection<Parameter<*>>) {
    Parameters.apply(this, query, parameters).use { it.executeUpdate() }
}
