package solutions.bitbadger.documents.common

/**
 * A field and its comparison
 *
 * @property name The name of the field in the JSON document
 * @property comparison The comparison to apply against the field
 * @property parameterName The name of the parameter to use in the query (optional, generated if missing)
 * @property qualifier A table qualifier to use to address the `data` field (useful for multi-table queries)
 */
class Field<T>(
    val name: String,
    val comparison: Comparison<T>,
    val parameterName: String? = null,
    val qualifier: String? = null) {

    /**
     * Specify the parameter name for the field
     *
     * @param paramName The parameter name to use for this field
     * @return A new `Field` with the parameter name specified
     */
    fun withParameterName(paramName: String): Field<T> =
        Field(name, comparison, paramName, qualifier)

    /**
     * Specify a qualifier (alias) for the document table
     *
     * @param alias The table alias for this field
     * @return A new `Field` with the table qualifier specified
     */
    fun withQualifier(alias: String): Field<T> =
        Field(name, comparison, parameterName, alias)

    /**
     * Get the path for this field
     *
     * @param dialect The SQL dialect to use for the path to the JSON field
     * @param format Whether the value should be retrieved as JSON or SQL (optional, default SQL)
     * @return The path for the field
     */
    fun path(dialect: Dialect, format: FieldFormat = FieldFormat.SQL): String =
        (if (qualifier == null) "" else "${qualifier}.") + nameToPath(name, dialect, format)

    companion object {
        /**
         * Create a field equality comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> equal(name: String, value: T): Field<T> =
            Field<T>(name, Comparison(Op.EQUAL, value))

        /**
         * Create a field greater-than comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> greater(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.GREATER, value))

        /**
         * Create a field greater-than-or-equal-to comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> greaterOrEqual(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.GREATER_OR_EQUAL, value))

        /**
         * Create a field less-than comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> less(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.LESS, value))

        /**
         * Create a field less-than-or-equal-to comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> lessOrEqual(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.LESS_OR_EQUAL, value))

        /**
         * Create a field inequality comparison
         *
         * @param name The name of the field to be compared
         * @param value The value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> notEqual(name: String, value: T): Field<T> =
            Field(name, Comparison(Op.NOT_EQUAL, value))

        /**
         * Create a field range comparison
         *
         * @param name The name of the field to be compared
         * @param minValue The lower value for the comparison
         * @param maxValue The upper value for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> between(name: String, minValue: T, maxValue: T): Field<Pair<T, T>> =
            Field(name, Comparison(Op.BETWEEN, Pair(minValue, maxValue)))

        /**
         * Create a field where any values match (SQL `IN`)
         *
         * @param name The name of the field to be compared
         * @param values The values for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> any(name: String, values: List<T>): Field<List<T>> =
            Field(name, Comparison(Op.IN, values))

        /**
         * Create a field where values should exist in a document's array
         *
         * @param name The name of the field to be compared
         * @param tableName The name of the document table
         * @param values The values for the comparison
         * @return A `Field` with the given comparison
         */
        fun <T> inArray(name: String, tableName: String, values: List<T>): Field<Pair<String, List<T>>> =
            Field(name, Comparison(Op.IN_ARRAY, Pair(tableName, values)))

        fun exists(name: String): Field<String> =
            Field(name, Comparison(Op.EXISTS, ""))

        fun notExists(name: String): Field<String> =
            Field(name, Comparison(Op.NOT_EXISTS, ""))

        fun named(name: String): Field<String> =
            Field(name, Comparison(Op.EQUAL, ""))

        fun nameToPath(name: String, dialect: Dialect, format: FieldFormat): String {
            val path = StringBuilder("data")
            val extra = if (format == FieldFormat.SQL) ">" else ""
            if (name.indexOf('.') > -1) {
                if (dialect == Dialect.POSTGRESQL) {
                    path.append("#>", extra, "'{", name.replace('.', ','), "}'")
                } else {
                    val names = name.split('.').toMutableList()
                    val last = names.removeLast()
                    names.forEach { path.append("->'", it, "'") }
                    path.append("->", extra, "'", last, "'")
                }
            } else {
                path.append("->", extra, "'", name, "'")
            }
            return path.toString()
        }
    }
}