package solutions.bitbadger.documents.sqlite

import solutions.bitbadger.documents.common.*
import solutions.bitbadger.documents.common.Configuration as BaseConfig;
import solutions.bitbadger.documents.common.Query

/**
 * Queries with specific syntax in SQLite
 */
object Query {

    /**
     * Create a `WHERE` clause fragment to implement a comparison on fields in a JSON document
     *
     * @param howMatched How the fields should be matched
     * @param fields The fields for the comparisons
     * @return A `WHERE` clause implementing the comparisons for the given fields
     */
    fun whereByFields(howMatched: FieldMatch, fields: Collection<Field<*>>): String {
        val name = ParameterName()
        return fields.joinToString(" ${howMatched.sql} ") {
            val comp = it.comparison
            when (comp.op) {
                Op.EXISTS, Op.NOT_EXISTS -> {
                    "${it.path(Dialect.SQLITE, FieldFormat.SQL)} ${it.comparison.op.sql}"
                }
                Op.BETWEEN -> {
                    val p = name.derive(it.parameterName)
                    "${it.path(Dialect.SQLITE, FieldFormat.SQL)} ${comp.op.sql} ${p}min AND ${p}max"
                }
                Op.IN -> {
                    val p = name.derive(it.parameterName)
                    val values = comp.value as Collection<*>
                    val paramNames = values.indices.joinToString(", ") { idx -> "${p}_$idx" }
                    "${it.path(Dialect.SQLITE, FieldFormat.SQL)} ${comp.op.sql} ($paramNames)"
                }
                Op.IN_ARRAY -> {
                    val p = name.derive(it.parameterName)
                    @Suppress("UNCHECKED_CAST")
                    val tableAndValues = comp.value as? Pair<String, Collection<*>>
                        ?: throw IllegalArgumentException("InArray field invalid")
                    val (table, values) = tableAndValues
                    val paramNames = values.indices.joinToString(", ") { idx -> "${p}_$idx" }
                    "EXISTS (SELECT 1 FROM json_each($table.data, '$.${it.name}') WHERE value IN ($paramNames)"
                }
                else -> {
                    "${it.path(Dialect.SQLITE, FieldFormat.SQL)} ${comp.op.sql} ${name.derive(it.parameterName)}"
                }
            }
        }
    }

    /**
     * Create a `WHERE` clause fragment to implement an ID-based query
     *
     * @param docId The ID of the document
     * @return A `WHERE` clause fragment identifying a document by its ID
     */
    fun <TKey> whereById(docId: TKey): String =
        whereByFields(FieldMatch.ANY,
            listOf(Field.equal(BaseConfig.idField, docId).withParameterName(":id")))

    /**
     * Create an `UPDATE` statement to patch documents
     *
     * @param tableName The table to be updated
     * @return A query to patch documents
     */
    fun patch(tableName: String): String =
        "UPDATE $tableName SET data = json_patch(data, json(:data))"

    // TODO: fun removeFields(tableName: String, fields: Collection<String>): String

    /**
     * Create a query by a document's ID
     *
     * @param statement The SQL statement to be run against a document by its ID
     * @param docId The ID of the document targeted
     * @returns A query addressing a document by its ID
     */
    fun <TKey> byId(statement: String, docId: TKey): String =
        Query.statementWhere(statement, whereById(docId))

    /**
     * Create a query on JSON fields
     *
     * @param statement The SQL statement to be run against matching fields
     * @param howMatched Whether to match any or all of the field conditions
     * @param fields The field conditions to be matched
     * @return A query addressing documents by field matching conditions
     */
    fun byFields(statement: String, howMatched: FieldMatch, fields: Collection<Field<*>>): String =
        Query.statementWhere(statement, whereByFields(howMatched, fields))

    object Definition {

        /**
         * SQL statement to create a document table
         *
         * @param tableName The name of the table (may include schema)
         * @return A query to create the table if it does not exist
         */
        fun ensureTable(tableName: String): String =
            Query.Definition.ensureTableFor(tableName, "TEXT")
    }
}
