package solutions.bitbadger.documents.sqlite

import org.junit.jupiter.api.DisplayName
import solutions.bitbadger.documents.common.Field
import solutions.bitbadger.documents.common.FieldMatch
import kotlin.test.Test
import kotlin.test.assertEquals

class QueryTest {

    @Test
    @DisplayName("whereByFields generates for a single field with logical operator")
    fun whereByFieldSingleLogical() {
        assertEquals("data->>'theField' > :test",
            Query.whereByFields(FieldMatch.ANY, listOf(Field.greater("theField", 0).withParameterName(":test"))),
            "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereByFields generates for a single field with existence operator")
    fun whereByFieldSingleExistence() {
        assertEquals("data->>'thatField' IS NULL",
            Query.whereByFields(FieldMatch.ANY, listOf(Field.notExists("thatField"))), "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereByFields generates for a single field with between operator")
    fun whereByFieldSingleBetween() {
        assertEquals("data->>'aField' BETWEEN :rangemin AND :rangemax",
            Query.whereByFields(FieldMatch.ALL, listOf(Field.between("aField", 50, 99).withParameterName(":range"))),
            "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereByFields generates for all multiple fields with logical operator")
    fun whereByFieldAllMultipleLogical() {
        assertEquals("data->>'theFirst' = :field0 AND data->>'numberTwo' = :field1",
            Query.whereByFields(FieldMatch.ALL, listOf(Field.equal("theFirst", "1"), Field.equal("numberTwo", "2"))),
            "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereByFields generates for any multiple fields with existence operator")
    fun whereByFieldAnyMultipleExistence() {
        assertEquals("data->>'thatField' IS NULL OR data->>'thisField' >= :field0",
            Query.whereByFields(FieldMatch.ANY,
                listOf(Field.notExists("thatField"), Field.greaterOrEqual("thisField", 18))),
            "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereByFields generates for an In comparison")
    fun whereByFieldIn() {
        assertEquals("data->>'this' IN (:field0_0, :field0_1, :field0_2)",
            Query.whereByFields(FieldMatch.ALL, listOf(Field.any("this", listOf("a", "b", "c")))),
            "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereByFields generates for an InArray comparison")
    fun whereByFieldInArray() {
        assertEquals("EXISTS (SELECT 1 FROM json_each(the_table.data, '$.this') WHERE value IN (:field0_0, :field0_1)",
            Query.whereByFields(FieldMatch.ALL, listOf(Field.inArray("this", "the_table", listOf("a", "b")))),
            "WHERE clause not correct")
    }

    @Test
    @DisplayName("whereById generates correctly")
    fun whereById() {
        assertEquals("data->>'id' = :id", Query.whereById("abc"), "WHERE clause not correct")
    }

    @Test
    @DisplayName("patch generates the correct query")
    fun patch() {
        assertEquals("UPDATE my_table SET data = json_patch(data, json(:data))", Query.patch("my_table"),
            "Query not correct")
    }

    @Test
    @DisplayName("byId generates the correct query")
    fun byId() {
        assertEquals("test WHERE data->>'id' = :id", Query.byId("test", "14"), "By-ID query not correct")
    }

    @Test
    @DisplayName("byFields generates the correct query")
    fun byFields() {
        assertEquals("unit WHERE data->>'That' > :field0",
            Query.byFields("unit", FieldMatch.ANY, listOf(Field.greater("That", 14))), "By-fields query not correct")
    }

    @Test
    @DisplayName("Definition.ensureTable generates the correct query")
    fun ensureTable() {
        assertEquals("CREATE TABLE IF NOT EXISTS tbl (data TEXT NOT NULL)", Query.Definition.ensureTable("tbl"),
            "CREATE TABLE statement not correct")
    }
}
